//
//  FastSwitchingAccountView.swift
//  Tusker
//
//  Created by Shadowfacts on 11/4/20.
//  Copyright © 2020 Shadowfacts. All rights reserved.
//

import UIKit

class FastSwitchingAccountView: UIView {
    
    private static let selectedColor = UIColor { (traits) in
        if traits.userInterfaceStyle == .dark {
            return UIColor(hue: 211 / 360, saturation: 85 / 100, brightness: 100 / 100, alpha: 1)
        } else {
            return UIColor(hue: 211 / 360, saturation: 70 / 100, brightness: 100 / 100, alpha: 1)
        }
    }
    private static let currentColor = UIColor { (traits) in
        if traits.userInterfaceStyle == .dark {
            return UIColor(hue: 211 / 360, saturation: 85 / 100, brightness: 85 / 100, alpha: 1)
        } else {
            return UIColor(hue: 211 / 360, saturation: 50 / 100, brightness: 100 / 100, alpha: 1)
        }
    }
    var isSelected = false {
        didSet {
            updateLabelColors()
        }
    }
    var isCurrent = false {
        didSet {
            updateLabelColors()
        }
    }
    
    private let orientation: FastAccountSwitcherViewController.ItemOrientation
    
    private let usernameLabel = UILabel()
    private let instanceLabel = UILabel()
    private let avatarImageView = UIImageView()
    
    private var avatarRequest: ImageCache.Request?
    
    init(account: LocalData.UserAccountInfo, orientation: FastAccountSwitcherViewController.ItemOrientation) {
        self.orientation = orientation
        super.init(frame: .zero)
        commonInit()
        setupAccount(account: account)
    }
    
    init(orientation: FastAccountSwitcherViewController.ItemOrientation) {
        self.orientation = orientation
        super.init(frame: .zero)
        commonInit()
        setupPlaceholder()
    }
    
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    private func commonInit() {
        usernameLabel.textColor = .white
        usernameLabel.font = UIFont(descriptor: .preferredFontDescriptor(withTextStyle: .headline), size: 0)
        
        instanceLabel.textColor = .white
        instanceLabel.font = UIFont(descriptor: .preferredFontDescriptor(withTextStyle: .subheadline), size: 0)
        
        let stackView = UIStackView(arrangedSubviews: [
            usernameLabel,
            instanceLabel
        ])
        stackView.translatesAutoresizingMaskIntoConstraints = false
        stackView.axis = .vertical
        addSubview(stackView)
        
        avatarImageView.translatesAutoresizingMaskIntoConstraints = false
        avatarImageView.layer.masksToBounds = true
        avatarImageView.layer.cornerRadius = Preferences.shared.avatarStyle.cornerRadiusFraction * 40
        avatarImageView.image = UIImage(systemName: Preferences.shared.avatarStyle == .circle ? "person.crop.circle" : "person.crop.square")
        avatarImageView.contentMode = .scaleAspectFill
        addSubview(avatarImageView)
        
        NSLayoutConstraint.activate([
            avatarImageView.topAnchor.constraint(equalTo: topAnchor, constant: 8),
            avatarImageView.bottomAnchor.constraint(equalTo: bottomAnchor, constant: -8),
            avatarImageView.widthAnchor.constraint(equalToConstant: 40),
            avatarImageView.heightAnchor.constraint(equalTo: avatarImageView.widthAnchor),
            
            stackView.centerYAnchor.constraint(equalTo: avatarImageView.centerYAnchor),
        ])
        
        switch orientation {
        case .iconsLeading:
            stackView.alignment = .leading
            NSLayoutConstraint.activate([
                avatarImageView.leadingAnchor.constraint(equalTo: leadingAnchor),
                stackView.leadingAnchor.constraint(equalTo: avatarImageView.trailingAnchor, constant: 8),
                stackView.trailingAnchor.constraint(equalTo: trailingAnchor),
            ])
            
        case .iconsTrailing:
            stackView.alignment = .trailing
            NSLayoutConstraint.activate([
                avatarImageView.trailingAnchor.constraint(equalTo: trailingAnchor),
                stackView.leadingAnchor.constraint(equalTo: leadingAnchor, constant: 8),
                stackView.trailingAnchor.constraint(equalTo: avatarImageView.leadingAnchor, constant: -8),
            ])
        }
        
        updateLabelColors()
    }
    
    private func setupAccount(account: LocalData.UserAccountInfo) {
        usernameLabel.text = account.username
        instanceLabel.text = account.instanceURL.host!
        let controller = MastodonController.getForAccount(account)
        controller.getOwnAccount { [weak self] (result) in
            guard let self = self,
                  case let .success(account) = result,
                  let avatar = account.avatar else { return }
            self.avatarRequest = ImageCache.avatars.get(avatar) { [weak self] (_, image) in
                guard let self = self, let image = image else { return }
                DispatchQueue.main.async {
                    self.avatarImageView.image = image
                }
            }
        }
    }
    
    private func setupPlaceholder() {
        usernameLabel.text = "Add Account"
        instanceLabel.isHidden = true
        avatarImageView.image = UIImage(systemName: "plus")
    }
    
    private func updateLabelColors() {
        let color: UIColor
        if isSelected {
            color = FastSwitchingAccountView.selectedColor
        } else if isCurrent {
            color = FastSwitchingAccountView.currentColor
        } else {
            color = .white
        }
        usernameLabel.textColor = color
        instanceLabel.textColor = color
        avatarImageView.tintColor = color
    }

}
