//
//  PreviewViewControllerProvider.swift
//  Tusker
//
//  Created by Shadowfacts on 10/10/18.
//  Copyright © 2018 Shadowfacts. All rights reserved.
//

import UIKit
import SafariServices
import Pachyderm
import WebURLFoundationExtras

protocol MenuActionProvider: AnyObject {
    var navigationDelegate: TuskerNavigationDelegate? { get }
    var toastableViewController: ToastableViewController? { get }
}

protocol MenuPreviewProvider: AnyObject {
    typealias PreviewProviders = (content: UIContextMenuContentPreviewProvider, actions: () -> [UIMenuElement])
    
    func getPreviewProviders(for location: CGPoint, sourceViewController: UIViewController) -> PreviewProviders?
}

protocol CustomPreviewPresenting {
    func presentFromPreview(presenter: UIViewController)
}

extension MenuActionProvider where Self: TuskerNavigationDelegate {
    var navigationDelegate: TuskerNavigationDelegate? { self }
}

extension MenuActionProvider where Self: ToastableViewController {
    var toastableViewController: ToastableViewController? { self }
}

extension MenuActionProvider {
    
    private var mastodonController: MastodonController? { navigationDelegate?.apiController }
    
    func actionsForProfile(accountID: String, sourceView: UIView?) -> [UIMenuElement] {
        guard let mastodonController = mastodonController,
            let account = mastodonController.persistentContainer.account(for: accountID) else { return [] }
        
        guard let loggedInAccountID = mastodonController.accountInfo?.id else {
            // user is logged out
            return [
                openInSafariAction(url: account.url),
                createAction(identifier: "share", title: "Share", systemImageName: "square.and.arrow.up", handler: { [weak self, weak sourceView] (_) in
                    guard let self = self else { return }
                    self.navigationDelegate?.showMoreOptions(forAccount: accountID, sourceView: sourceView)
                })
            ]
        }
        
        let actionsSection: [UIMenuElement] = [
            createAction(identifier: "sendmessage", title: "Send Message", systemImageName: "envelope", handler: { [weak self] (_) in
                guard let self = self else { return }
                let draft = self.mastodonController!.createDraft(mentioningAcct: account.acct)
                draft.visibility = .direct
                self.navigationDelegate?.compose(editing: draft)
            }),
            UIDeferredMenuElement.uncached({ (elementHandler) in
                Task { @MainActor in
                    if let action = await self.followAction(for: accountID, mastodonController: mastodonController) {
                        elementHandler([action])
                    } else {
                        elementHandler([])
                    }
                }
            })
        ]
        
        var shareSection = [
            openInSafariAction(url: account.url),
            createAction(identifier: "share", title: "Share", systemImageName: "square.and.arrow.up", handler: { [weak self, weak sourceView] (_) in
                guard let self = self else { return }
                self.navigationDelegate?.showMoreOptions(forAccount: accountID, sourceView: sourceView)
            })
        ]
        
        addOpenInNewWindow(actions: &shareSection, activity: UserActivityManager.showProfileActivity(id: accountID, accountID: loggedInAccountID))
        
        return [
            UIMenu(title: "", image: nil, identifier: nil, options: [.displayInline], children: shareSection),
            UIMenu(title: "", image: nil, identifier: nil, options: [.displayInline], children: actionsSection),
        ]
    }
    
    func actionsForURL(_ url: URL, sourceView: UIView?) -> [UIAction] {
        return [
            openInSafariAction(url: url),
            createAction(identifier: "share", title: "Share", systemImageName: "square.and.arrow.up", handler: { [weak self, weak sourceView] (_) in
                guard let self = self else { return }
                self.navigationDelegate?.showMoreOptions(forURL: url, sourceView: sourceView)
            })
        ]
    }
    
    func actionsForHashtag(_ hashtag: Hashtag, sourceView: UIView?) -> [UIMenuElement] {
        let actionsSection: [UIMenuElement]
        if let mastodonController = mastodonController,
           mastodonController.loggedIn {
            let context = mastodonController.persistentContainer.viewContext
            let existing = try? context.fetch(SavedHashtag.fetchRequest(name: hashtag.name)).first
            actionsSection = [
                createAction(identifier: "save", title: existing != nil ? "Unsave Hashtag" : "Save Hashtag", systemImageName: "number", handler: { (_) in
                    if let existing = existing {
                        context.delete(existing)
                    } else {
                        _ = SavedHashtag(hashtag: hashtag, context: context)
                    }
                    try! context.save()
                })
            ]
        } else {
            actionsSection = []
        }
        
        let shareSection: [UIMenuElement]
        if let url = URL(hashtag.url) {
            shareSection = actionsForURL(url, sourceView: sourceView)
        } else {
            shareSection = []
        }
        
        return [
            UIMenu(title: "", image: nil, identifier: nil, options: [.displayInline], children: shareSection),
            UIMenu(title: "", image: nil, identifier: nil, options: [.displayInline], children: actionsSection),
        ]
    }
    
    func actionsForStatus(_ status: StatusMO, sourceView: UIView?, includeReply: Bool = true) -> [UIMenuElement] {
        guard let mastodonController = mastodonController else { return [] }
        
        guard let accountID = mastodonController.accountInfo?.id else {
            // user is logged out
            return [
                openInSafariAction(url: status.url!),
                createAction(identifier: "share", title: "Share", systemImageName: "square.and.arrow.up", handler: { [weak self, weak sourceView] (_) in
                    guard let self = self else { return }
                    self.navigationDelegate?.showMoreOptions(forStatus: status.id, sourceView: sourceView)
                })
            ]
        }
        let bookmarked = status.bookmarked ?? false
        
        var toggleableSection = [
            createAction(identifier: "bookmark", title: bookmarked ? "Unbookmark" : "Bookmark", systemImageName: bookmarked ? "bookmark.fill" : "bookmark", handler: { [weak self] (_) in
                guard let self = self else { return }
                let request = (bookmarked ? Status.unbookmark : Status.bookmark)(status.id)
                self.mastodonController?.run(request) { (response) in
                    switch response {
                    case .success(let status, _):
                        self.mastodonController?.persistentContainer.addOrUpdate(status: status)
                        
                    case .failure(let error):
                        if let toastable = self.toastableViewController {
                            let config = ToastConfiguration(from: error, with: "Error \(bookmarked ? "Unb" : "B")ookmarking", in: toastable, retryAction: nil)
                            DispatchQueue.main.async {
                                toastable.showToast(configuration: config, animated: true)
                            }
                        }
                    }
                }
            }),
        ]
        
        if #available(iOS 16.0, *) {
            let favorited = status.favourited
            // TODO: move this color into an asset catalog or something
            var favImage = UIImage(systemName: favorited ? "star.fill" : "star")!
            if favorited {
                favImage = favImage.withTintColor(UIColor(displayP3Red: 1, green: 0.8, blue: 0, alpha: 1), renderingMode: .alwaysOriginal)
            }
            toggleableSection.insert(createAction(identifier: "favorite", title: favorited ? "Unfavorite" : "Favorite", image: favImage, handler: { [weak self] _ in
                guard let self = self else { return }
                let request = (favorited ? Status.favourite : Status.unfavourite)(status.id)
                self.mastodonController?.run(request, completion: { response in
                    switch response {
                    case .success(let status, _):
                        self.mastodonController?.persistentContainer.addOrUpdate(status: status)
                    case .failure(let error):
                        if let toastable = self.toastableViewController {
                            let config = ToastConfiguration(from: error, with: "Error \(favorited ? "Unf" : "F")avoriting", in: toastable, retryAction: nil)
                            DispatchQueue.main.async {
                                toastable.showToast(configuration: config, animated: true)
                            }
                        }
                    }
                })
            }), at: 0)
            
            let reblogged = status.reblogged
            var reblogImage = UIImage(systemName: "repeat")!
            if reblogged {
                reblogImage = reblogImage.withTintColor(UIColor(displayP3Red: 1, green: 0.8, blue: 0, alpha: 1), renderingMode: .alwaysOriginal)
            }
            toggleableSection.insert(createAction(identifier: "reblog", title: reblogged ? "Unreblog" : "Reblog", image: reblogImage, handler: { [weak self] _ in
                guard let self = self else { return }
                let request = (reblogged ? Status.reblog : Status.unreblog)(status.id)
                self.mastodonController?.run(request, completion: { response in
                    switch response {
                    case .success(let status, _):
                        self.mastodonController?.persistentContainer.addOrUpdate(status: status)
                    case .failure(let error):
                        if let toastable = self.toastableViewController {
                            let config = ToastConfiguration(from: error, with: "Error \(reblogged ? "Unr" : "R")eblogging", in: toastable, retryAction: nil)
                            DispatchQueue.main.async {
                                toastable.showToast(configuration: config, animated: true)
                            }
                        }
                    }
                })
            }), at: 1)
        }
        
        var actionsSection: [UIAction] = []
        
        if includeReply {
            actionsSection.insert(createAction(identifier: "reply", title: "Reply", systemImageName: "arrowshape.turn.up.left", handler: { [weak self] (_) in
                guard let self = self else { return }
                self.navigationDelegate?.compose(inReplyToID: status.id)
            }), at: 0)
        }
        
        if let account = mastodonController.account {
            // only allow muting conversations that either current user posted or is participating in (technically, is mentioned, since that's the best we can do)
            if status.account.id == account.id || status.mentions.contains(where: { $0.id == account.id }) {
                let muted = status.muted
                toggleableSection.append(createAction(identifier: "mute", title: muted ? "Unmute Conversation" : "Mute Conversation", systemImageName: muted ? "speaker" : "speaker.slash", handler: { [weak self] (_) in
                    guard let self = self else { return }
                    let request = (muted ? Status.unmuteConversation : Status.muteConversation)(status.id)
                    self.mastodonController?.run(request) { (response) in
                        switch response {
                        case .success(let status, _):
                            self.mastodonController?.persistentContainer.addOrUpdate(status: status)
                        case .failure(let error):
                            if let toastable = self.toastableViewController {
                                let config = ToastConfiguration(from: error, with: "Error \(muted ? "Unm" : "M")uting", in: toastable, retryAction: nil)
                                DispatchQueue.main.async {
                                    toastable.showToast(configuration: config, animated: true)
                                }
                            }
                        }
                    }
                }))
            }
            
            // only allowing pinning user's own statuses
            if account.id == status.account.id,
               mastodonController.instanceFeatures.profilePinnedStatuses {
                let pinned = status.pinned ?? false
                toggleableSection.append(createAction(identifier: "pin", title: pinned ? "Unpin from Profile" : "Pin to Profile", systemImageName: pinned ? "pin.slash" : "pin", handler: { [weak self] (_) in
                    guard let self = self else { return }
                    let request = (pinned ? Status.unpin : Status.pin)(status.id)
                    self.mastodonController?.run(request, completion: { [weak self] (response) in
                        guard let self = self else { return }
                        switch response {
                        case .success(let status, _):
                            self.mastodonController?.persistentContainer.addOrUpdate(status: status)
                        case .failure(let error):
                            if let toastable = self.toastableViewController {
                                let config = ToastConfiguration(from: error, with: "Error \(pinned ? "Unp" :"P")inning", in: toastable, retryAction: nil)
                                DispatchQueue.main.async {
                                    toastable.showToast(configuration: config, animated: true)
                                }
                            }
                        }
                    })
                }))
            }
        }
        
        if status.poll != nil {
            actionsSection.insert(createAction(identifier: "refresh", title: "Refresh Poll", systemImageName: "arrow.clockwise", handler: { [weak self] (_) in
                guard let mastodonController = self?.mastodonController else { return }
                let request = Client.getStatus(id: status.id)
                mastodonController.run(request, completion: { (response) in
                    switch response {
                    case .success(let status, _):
                        // todo: this shouldn't really use the viewContext, but for some reason saving the
                        // backgroundContext with the new version of the status isn't updating the viewContext
                        DispatchQueue.main.async {
                            mastodonController.persistentContainer.addOrUpdate(status: status, context: mastodonController.persistentContainer.viewContext)
                        }
                    case .failure(let error):
                        if let toastable = self?.toastableViewController {
                            let config = ToastConfiguration(from: error, with: "Error Refreshing Poll", in: toastable, retryAction: nil)
                            DispatchQueue.main.async {
                                toastable.showToast(configuration: config, animated: true)
                            }
                        }
                    }
                })
            }), at: 0)
        }
        
        var shareSection = [
            openInSafariAction(url: status.url!),
            createAction(identifier: "share", title: "Share", systemImageName: "square.and.arrow.up", handler: { [weak self, weak sourceView] (_) in
                guard let self = self else { return }
                self.navigationDelegate?.showMoreOptions(forStatus: status.id, sourceView: sourceView)
            }),
        ]
        
        addOpenInNewWindow(actions: &shareSection, activity: UserActivityManager.showConversationActivity(mainStatusID: status.id, accountID: accountID))
        
        if #available(iOS 16.0, *) {
            return [
                UIMenu(options: .displayInline, preferredElementSize: .medium, children: toggleableSection + actionsSection),
                UIMenu(options: .displayInline, children: shareSection),
            ]
        } else {
            return [
                UIMenu(options: .displayInline, children: shareSection),
                UIMenu(options: .displayInline, children: toggleableSection),
                UIMenu(options: .displayInline, children: actionsSection),
            ]
        }
    }
    
    func actionsForTrendingLink(card: Card) -> [UIMenuElement] {
        guard let url = URL(card.url) else {
            return []
        }
        return [
            openInSafariAction(url: url),
            createAction(identifier: "postlink", title: "Post this Link", systemImageName: "square.and.pencil", handler: { [weak self] _ in
                guard let self = self else { return }
                let draft = self.mastodonController!.createDraft()
                let title = card.title.trimmingCharacters(in: .whitespacesAndNewlines)
                if !title.isEmpty {
                    draft.text += title
                    draft.text += ":\n"
                }
                draft.text += url.absoluteString
                // prevents the draft from being saved automatically until the user makes a change
                // also prevents it from being posted without being changed
                draft.initialText = draft.text
                self.navigationDelegate?.compose(editing: draft)
            })
        ]
    }

    private func createAction(identifier: String, title: String, systemImageName: String?, handler: @escaping UIActionHandler) -> UIAction {
        let image: UIImage?
        if let name = systemImageName {
            image = UIImage(systemName: name)
        } else {
            image = nil
        }
        return createAction(identifier: identifier, title: title, image: image, handler: handler)
    }
    
    private func createAction(identifier: String, title: String, image: UIImage?, handler: @escaping UIActionHandler) -> UIAction {
        return UIAction(title: title, image: image, identifier: UIAction.Identifier(identifier), discoverabilityTitle: nil, attributes: [], state: .off, handler: handler)
    }
    
    private func openInSafariAction(url: URL) -> UIAction {
        return createAction(identifier: "openinsafari", title: "Open in Safari", systemImageName: "safari", handler: { [weak self] (_) in
            self?.navigationDelegate?.selected(url: url, allowUniversalLinks: false)
        })
    }
    
    private func addOpenInNewWindow(actions: inout [UIAction], activity: @escaping @autoclosure () -> NSUserActivity) {
        let options = UIWindowScene.ActivationRequestOptions()
        options.preferredPresentationStyle = .automatic
        actions.append(UIWindowScene.ActivationAction { (_) in
            let activity = activity()
            activity.displaysAuxiliaryScene = true
            return .init(userActivity: activity, options: options, preview: nil)
        })
    }
    
    private func followAction(for accountID: String, mastodonController: MastodonController) async -> UIMenuElement? {
        guard let ownAccount = try? await mastodonController.getOwnAccount(),
              accountID != ownAccount.id else {
            return nil
        }
        let request = Client.getRelationships(accounts: [accountID])
        guard let (relationships, _) = try? await mastodonController.run(request),
              let relationship = relationships.first else {
            return nil
        }
        let following = relationship.following
        return createAction(identifier: "follow", title: following ? "Unfollow" : "Follow", systemImageName: following ? "person.badge.minus" : "person.badge.plus") { _ in
            let request = (following ? Account.unfollow : Account.follow)(accountID)
            mastodonController.run(request) { response in
                switch response {
                case .failure(let error):
                    if let toastable = self.toastableViewController {
                        let config = ToastConfiguration(from: error, with: "Error \(following ? "Unf" : "F")ollowing", in: toastable, retryAction: nil)
                        DispatchQueue.main.async {
                            toastable.showToast(configuration: config, animated: true)
                        }
                    }
                case .success(let relationship, _):
                    mastodonController.persistentContainer.addOrUpdate(relationship: relationship)
                }
            }
        }
    }
    
}

extension LargeImageViewController: CustomPreviewPresenting {
    func presentFromPreview(presenter: UIViewController) {
        presenter.present(self, animated: true)
    }
}

extension GalleryViewController: CustomPreviewPresenting {
    func presentFromPreview(presenter: UIViewController) {
        presenter.present(self, animated: true)
    }
}

extension SFSafariViewController: CustomPreviewPresenting {
    func presentFromPreview(presenter: UIViewController) {
        presenter.present(self, animated: true)
    }
}
